<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Your Verification Token</title>
</head>
<body>
    <h2>Use this verification code: {{ callback_token }}</h2>
</body>
</html>